@@ -146,12 +146,15 @@ diff_blobs(struct got_diff_line **lines, size_t *nlines,
 	off_t outoff = 0;
 	int n;
 
-	if (lines && *lines && *nlines > 0)
-		outoff = (*lines)[*nlines - 1].offset;
-	else if (lines) {
-		err = add_line_metadata(lines, nlines, 0, GOT_DIFF_LINE_NONE);
-		if (err)
-			goto done;
+	if (lines && *lines) {
+		if (*nlines > 0)
+			outoff = (*lines)[*nlines - 1].offset;
+		else {
+			err = add_line_metadata(lines, nlines,
+			    0, GOT_DIFF_LINE_NONE);
+			if (err != NULL)
+				goto done;
+		}
 	}
 
 	if (resultp)
@@ -218,7 +221,7 @@ diff_blobs(struct got_diff_line **lines, size_t *nlines,
 		if (n < 0)
 			goto done;
 		outoff += n;
-		if (lines) {
+		if (lines && *lines) {
 			err = add_line_metadata(lines, nlines, outoff,
 			    GOT_DIFF_LINE_BLOB_MIN);
 			if (err)
@@ -230,7 +233,7 @@ diff_blobs(struct got_diff_line **lines, size_t *nlines,
 		if (n < 0)
 			goto done;
 		outoff += n;
-		if (lines) {
+		if (lines && *lines) {
 			err = add_line_metadata(lines, nlines, outoff,
 			    GOT_DIFF_LINE_BLOB_PLUS);
 			if (err)
@@ -378,8 +381,15 @@ diff_blob_file(struct got_diffreg_result **resultp,
 
 	err = got_diffreg(&result, f1, f2, diff_algo, ignore_whitespace,
 	    force_text_diff);
-	if (err)
+	if (err) {
+		char msg[GOT_ERR_MAX_MSG_SIZE];
+		if (snprintf(msg, sizeof(msg), "%s vs %s: %s",
+		             label1 ? label1 : idstr1,
+		             f2_exists ? label2 : "/dev/null", err->msg) >= 0) {
+			err = got_error_msg(err->code, msg);
+		}
 		goto done;
+	}
 
 	if (outfile) {
 		err = got_diffreg_output(NULL, NULL, result,
@@ -669,11 +679,50 @@ done:
 }
 
 static const struct got_error *
-diff_kind_mismatch(struct got_object_id *id1, struct got_object_id *id2,
+diff_kind_mismatch(struct got_tree_entry *te1, struct got_tree_entry *te2,
+    FILE *f1, FILE *f2, int fd1, int fd2,
     const char *label1, const char *label2, struct got_repository *repo,
-    got_diff_blob_cb cb, void *cb_arg)
+    got_diff_blob_cb cb, void *cb_arg, int diff_content)
 {
-	/* XXX TODO */
+	const struct got_error *err = NULL;
+
+	/* 
+	 * Handle files changing into directories and vice-versa.
+	 * Disregard edge cases with FIFOs, device nodes, etc for now.
+	 */
+	if (!S_ISDIR(te1->mode) && S_ISDIR(te2->mode)) {
+		if (S_ISREG(te1->mode)) {
+			if (diff_content) {
+				err = diff_deleted_blob(&te1->id, f1, fd1,
+				    f2, label1, te1->mode, repo, cb, cb_arg);
+			} else {
+				err = cb(cb_arg, NULL, NULL, NULL, NULL,
+				    &te1->id, NULL, label1, NULL,
+				    te1->mode, 0, repo);
+			}
+			if (err)
+				return err;
+		}
+		return diff_added_tree(&te2->id, f1, f2, fd2, label2,
+		    repo, cb, cb_arg, diff_content);
+	} else if (S_ISDIR(te1->mode) && !S_ISDIR(te2->mode)) {
+		err = diff_deleted_tree(&te1->id, f1, fd1, f2,
+		    label1, repo, cb, cb_arg, diff_content);
+		if (err)
+			return err;
+		if (S_ISREG(te2->mode)) {
+			if (diff_content) {
+				err = diff_added_blob(&te2->id, f1, f2, fd2,
+				    label2, te2->mode, repo, cb, cb_arg);
+			} else {
+				err = cb(cb_arg, NULL, NULL, NULL, NULL, NULL,
+				    &te2->id, NULL, label2, 0, te2->mode, repo);
+			}
+			if (err)
+				return err;
+		}
+	}
+
 	return NULL;
 }
 
@@ -732,8 +781,8 @@ diff_entry_old_new(struct got_tree_entry *te1, struct got_tree_entry *te2,
 	if (id_match)
 		return NULL;
 
-	return diff_kind_mismatch(&te1->id, &te2->id, label1, label2, repo,
-	    cb, cb_arg);
+	return diff_kind_mismatch(te1, te2, f1, f2, fd1, fd2,
+	    label1, label2, repo, cb, cb_arg, diff_content);
 }
 
 static const struct got_error *
@@ -1059,15 +1108,18 @@ diff_paths(struct got_tree_object *tree1, struct got_tree_object *tree2,
 			got_object_blob_close(blob2);
 			blob2 = NULL;
 		}
-
-		err = got_object_tree_find_path(&id1, &mode1, repo, tree1,
-		    pe->path);
-		if (err && err->code != GOT_ERR_NO_TREE_ENTRY)
-			goto done;
-		err = got_object_tree_find_path(&id2, &mode2, repo, tree2,
-		    pe->path);
-		if (err && err->code != GOT_ERR_NO_TREE_ENTRY)
-			goto done;
+		if (tree1) {
+			err = got_object_tree_find_path(&id1, &mode1, repo,
+			    tree1, pe->path);
+			if (err && err->code != GOT_ERR_NO_TREE_ENTRY)
+				goto done;
+		}
+		if (tree2) {
+			err = got_object_tree_find_path(&id2, &mode2, repo,
+			    tree2, pe->path);
+			if (err && err->code != GOT_ERR_NO_TREE_ENTRY)
+				goto done;
+		}
 		if (id1 == NULL && id2 == NULL) {
 			err = got_error_path(pe->path, GOT_ERR_NO_TREE_ENTRY);
 			goto done;
