@@ -134,6 +134,9 @@ struct got_repository {
 
 	/* Settings read from got.conf. */
 	struct got_gotconfig *gotconfig;
+
+	/* cleanup lockfile */
+	struct got_lockfile *cleanup_lock;
 };
 
 const struct got_error*got_repo_cache_object(struct got_repository *,
@@ -181,3 +184,6 @@ const struct got_error *got_repo_read_gitconfig(int *, char **, char **,
 const struct got_error *got_repo_temp_fds_get(int *, int *,
     struct got_repository *);
 void got_repo_temp_fds_put(int, struct got_repository *);
+
+const struct got_error *got_repo_find_object_id(struct got_object_id *,
+    struct got_repository *);
