@@ -26,9 +26,47 @@
 #include <limits.h>
 
 #include "got_object.h"
+#include "got_error.h"
 
 #include "got_lib_hash.h"
 
+struct got_object_id *
+got_object_id_dup(struct got_object_id *id1)
+{
+	struct got_object_id *id2;
+
+	id2 = malloc(sizeof(*id2));
+	if (id2 == NULL)
+		return NULL;
+	memcpy(id2, id1, sizeof(*id2));
+	return id2;
+}
+
+int
+got_object_id_cmp(const struct got_object_id *id1,
+    const struct got_object_id *id2)
+{
+	return memcmp(id1->sha1, id2->sha1, SHA1_DIGEST_LENGTH);
+}
+
+const struct got_error *
+got_object_id_str(char **outbuf, struct got_object_id *id)
+{
+	static const size_t len = GOT_OBJECT_ID_HEX_MAXLEN;
+
+	*outbuf = malloc(len);
+	if (*outbuf == NULL)
+		return got_error_from_errno("malloc");
+
+	if (got_object_id_hex(id, *outbuf, len) == NULL) {
+		free(*outbuf);
+		*outbuf = NULL;
+		return got_error(GOT_ERR_BAD_OBJ_ID_STR);
+	}
+
+	return NULL;
+}
+
 int
 got_parse_xdigit(uint8_t *val, const char *hex)
 {
@@ -114,6 +152,12 @@ got_parse_hash_digest(uint8_t *digest, const char *line,
 	}
 }
 
+char *
+got_object_id_hex(struct got_object_id *id, char *buf, size_t len)
+{
+	return got_sha1_digest_to_str(id->sha1, buf, len);
+}
+
 int
 got_parse_object_id(struct got_object_id *id, const char *line,
     enum got_hash_algorithm algo)
@@ -161,7 +205,10 @@ void
 got_hash_final_object_id(struct got_hash *hash, struct got_object_id *id)
 {
 	memset(id, 0, sizeof(*id));
-	got_hash_final(hash, id->sha1);
+	if (hash->algo == GOT_HASH_SHA1)
+		SHA1Final(id->sha1, &hash->sha1_ctx);
+	else
+		abort();
 }
 
 int
