@@ -32,6 +32,7 @@
 #include "got_lib_delta.h"
 #include "got_lib_inflate.h"
 #include "got_lib_object.h"
+#include "got_lib_object_qid.h"
 #include "got_lib_object_idset.h"
 #include "got_lib_object_parse.h"
 
@@ -57,7 +58,7 @@ got_object_idset_alloc(void)
 	if (set == NULL)
 		return NULL;
 
-	set->ids = calloc(sizeof(set->ids[0]), GOT_OBJECT_IDSET_MIN_BUCKETS);
+	set->ids = calloc(GOT_OBJECT_IDSET_MIN_BUCKETS, sizeof(set->ids[0]));
 	if (set->ids == NULL) {
 		free(set);
 		return NULL;
