@@ -15,6 +15,8 @@
  * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  */
 
+#include "got_compat.h"
+
 #include <sys/types.h>
 #include <sys/queue.h>
 #include <sys/uio.h>
@@ -34,10 +36,6 @@
 #include <limits.h>
 #include <zlib.h>
 
-#if defined(__FreeBSD__)
-#include <unistd.h>
-#endif
-
 #include "got_error.h"
 #include "got_cancel.h"
 #include "got_object.h"
@@ -47,6 +45,7 @@
 
 #include "got_lib_deltify.h"
 #include "got_lib_delta.h"
+#include "got_lib_hash.h"
 #include "got_lib_object.h"
 #include "got_lib_object_idset.h"
 #include "got_lib_object_cache.h"
@@ -1419,14 +1418,14 @@ done:
 }
 
 static const struct got_error *
-hwrite(int fd, const void *buf, off_t len, SHA1_CTX *ctx)
+hwrite(int fd, const void *buf, off_t len, struct got_hash *ctx)
 {
-	SHA1Update(ctx, buf, len);
+	got_hash_update(ctx, buf, len);
 	return got_poll_write_full(fd, buf, len);
 }
 
 static const struct got_error *
-hcopy(FILE *fsrc, int fd_dst, off_t len, SHA1_CTX *ctx)
+hcopy(FILE *fsrc, int fd_dst, off_t len, struct got_hash *ctx)
 {
 	const struct got_error *err;
 	unsigned char buf[65536];
@@ -1438,7 +1437,7 @@ hcopy(FILE *fsrc, int fd_dst, off_t len, SHA1_CTX *ctx)
 		n = fread(buf, 1, copylen, fsrc);
 		if (n != copylen)
 			return got_ferror(fsrc, GOT_ERR_IO);
-		SHA1Update(ctx, buf, copylen);
+		got_hash_update(ctx, buf, copylen);
 		err = got_poll_write_full(fd_dst, buf, copylen);
 		if (err)
 			return err;
@@ -1450,12 +1449,12 @@ hcopy(FILE *fsrc, int fd_dst, off_t len, SHA1_CTX *ctx)
 
 static const struct got_error *
 hcopy_mmap(uint8_t *src, off_t src_offset, size_t src_size,
-    int fd, off_t len, SHA1_CTX *ctx)
+    int fd, off_t len, struct got_hash *ctx)
 {
 	if (src_offset + len > src_size)
 		return got_error(GOT_ERR_RANGE);
 
-	SHA1Update(ctx, src + src_offset, len);
+	got_hash_update(ctx, src + src_offset, len);
 	return got_poll_write_full(fd, src + src_offset, len);
 }
 
@@ -1550,8 +1549,8 @@ packoff(char *hdr, off_t off)
 }
 
 static const struct got_error *
-deltahdr(off_t *packfile_size, SHA1_CTX *ctx, int packfd, int force_refdelta,
-    struct got_pack_meta *m)
+deltahdr(off_t *packfile_size, struct got_hash *ctx, int packfd,
+    int force_refdelta, struct got_pack_meta *m)
 {
 	const struct got_error *err;
 	char buf[32];
@@ -1589,7 +1588,7 @@ deltahdr(off_t *packfile_size, SHA1_CTX *ctx, int packfd, int force_refdelta,
 static const struct got_error *
 write_packed_object(off_t *packfile_size, int packfd,
     FILE *delta_cache, uint8_t *delta_cache_map, size_t delta_cache_size,
-    struct got_pack_meta *m, int *outfd, SHA1_CTX *ctx,
+    struct got_pack_meta *m, int *outfd, struct got_hash *ctx,
     struct got_repository *repo, int force_refdelta)
 {
 	const struct got_error *err = NULL;
@@ -1599,8 +1598,8 @@ write_packed_object(off_t *packfile_size, int packfd,
 	struct got_raw_object *raw = NULL;
 	off_t outlen, delta_offset;
 
-	csum.output_sha1 = ctx;
-	csum.output_crc = NULL;
+	memset(&csum, 0, sizeof(csum));
+	csum.output_ctx = ctx;
 
 	if (m->reused_delta_offset)
 		delta_offset = m->reused_delta_offset;
@@ -1693,7 +1692,7 @@ genpack(uint8_t *pack_sha1, int packfd, struct got_pack *reuse_pack,
 {
 	const struct got_error *err = NULL;
 	int i;
-	SHA1_CTX ctx;
+	struct got_hash ctx;
 	struct got_pack_meta *m;
 	char buf[32];
 	off_t packfile_size = 0;
@@ -1703,7 +1702,7 @@ genpack(uint8_t *pack_sha1, int packfd, struct got_pack *reuse_pack,
 	size_t delta_cache_size = 0;
 	FILE *packfile = NULL;
 
-	SHA1Init(&ctx);
+	got_hash_init(&ctx, GOT_HASH_SHA1);
 
 #ifndef GOT_PACK_NO_MMAP
 	delta_cache_fd = dup(fileno(delta_cache));
@@ -1784,7 +1783,7 @@ genpack(uint8_t *pack_sha1, int packfd, struct got_pack *reuse_pack,
 			goto done;
 	}
 
-	SHA1Final(pack_sha1, &ctx);
+	got_hash_final(&ctx, pack_sha1);
 	err = got_poll_write_full(packfd, pack_sha1, SHA1_DIGEST_LENGTH);
 	if (err)
 		goto done;
@@ -1931,6 +1930,9 @@ got_pack_create(uint8_t *packsha1, int packfd, FILE *delta_cache,
 			goto done;
 	}
 
+	/* Pinned pack may have moved to different cache slot. */
+	reuse_pack = got_repo_get_pinned_pack(repo);
+
 	err = genpack(packsha1, packfd, reuse_pack, delta_cache, deltify.meta,
 	    deltify.nmeta, reuse.meta, reuse.nmeta, ncolored, nfound, ntrees,
 	    nours, repo, force_refdelta, progress_cb, progress_arg, rl,
