@@ -101,33 +101,6 @@ putbe32(char *b, uint32_t n)
 	b[3] = n >> 0;
 }
 
-static const struct got_error *
-get_obj_type_label(const char **label, int obj_type)
-{
-	const struct got_error *err = NULL;
-
-	switch (obj_type) {
-	case GOT_OBJ_TYPE_BLOB:
-		*label = GOT_OBJ_LABEL_BLOB;
-		break;
-	case GOT_OBJ_TYPE_TREE:
-		*label = GOT_OBJ_LABEL_TREE;
-		break;
-	case GOT_OBJ_TYPE_COMMIT:
-		*label = GOT_OBJ_LABEL_COMMIT;
-		break;
-	case GOT_OBJ_TYPE_TAG:
-		*label = GOT_OBJ_LABEL_TAG;
-		break;
-	default:
-		*label = NULL;
-		err = got_error(GOT_ERR_OBJ_TYPE);
-		break;
-	}
-
-	return err;
-}
-
 static const struct got_error *
 read_checksum(uint32_t *crc, struct got_hash *ctx, int fd, size_t len)
 {
@@ -238,7 +211,7 @@ read_packed_object(struct got_pack *pack, struct got_indexed_object *obj,
 		if (err)
 			break;
 		got_hash_init(&ctx, GOT_HASH_SHA1);
-		err = get_obj_type_label(&obj_label, obj->type);
+		err = got_object_type_label(&obj_label, obj->type);
 		if (err) {
 			free(data);
 			break;
@@ -271,12 +244,6 @@ read_packed_object(struct got_pack *pack, struct got_indexed_object *obj,
 				err = got_error(GOT_ERR_BAD_PACKFILE);
 				break;
 			}
-			if (mapoff + SHA1_DIGEST_LENGTH > SIZE_MAX) {
-				err = got_error_fmt(GOT_ERR_RANGE,
-				    "mapoff %lld would overflow size_t",
-				    (long long)mapoff + SHA1_DIGEST_LENGTH);
-				break;
-			}
 			memcpy(obj->delta.ref.ref_id.sha1, pack->map + mapoff,
 			    SHA1_DIGEST_LENGTH);
 			obj->crc = crc32(obj->crc, pack->map + mapoff,
@@ -332,6 +299,7 @@ read_packed_object(struct got_pack *pack, struct got_indexed_object *obj,
 				    "mapoff %lld would overflow size_t",
 				    (long long)mapoff
 				    + obj->delta.ofs.base_offsetlen);
+				break;
 			}
 
 			obj->crc = crc32(obj->crc, pack->map + mapoff,
@@ -437,7 +405,7 @@ resolve_deltified_object(struct got_pack *pack, struct got_packidx *packidx,
 	err = got_delta_chain_get_base_type(&base_obj_type, &deltas);
 	if (err)
 		goto done;
-	err = get_obj_type_label(&obj_label, base_obj_type);
+	err = got_object_type_label(&obj_label, base_obj_type);
 	if (err)
 		goto done;
 	if (asprintf(&header, "%s %zd", obj_label, len) == -1) {
