@@ -50,6 +50,12 @@
 #define nitems(_a)	(sizeof((_a)) / sizeof((_a)[0]))
 #endif
 
+/*
+ * Number of seconds until we give up trying to read more
+ * data from the client connection.
+ */
+static const int timeout = 60;
+
 static const struct got_capability read_capabilities[] = {
 	{ GOT_CAPA_AGENT, "got/" GOT_VERSION_STR },
 	{ GOT_CAPA_OFS_DELTA, NULL },
@@ -774,8 +780,7 @@ serve_read(int infd, int outfd, int gotd_sock, const char *repo_path,
 	enum protostate {
 		STATE_EXPECT_WANT,
 		STATE_EXPECT_MORE_WANT,
-		STATE_EXPECT_HAVE,
-		STATE_EXPECT_DONE,
+		STATE_EXPECT_HAVE_OR_DONE,
 		STATE_DONE,
 	};
 	enum protostate curstate = STATE_EXPECT_WANT;
@@ -792,14 +797,14 @@ serve_read(int infd, int outfd, int gotd_sock, const char *repo_path,
 	while (curstate != STATE_DONE) {
 		int n;
 		buf[0] = '\0';
-		err = got_pkt_readpkt(&n, infd, buf, sizeof(buf), chattygot);
+		err = got_pkt_readpkt(&n, infd, buf, sizeof(buf), chattygot,
+		    timeout);
 		if (err)
 			goto done;
 		if (n == 0) {
 			if (curstate != STATE_EXPECT_WANT &&
 			    curstate != STATE_EXPECT_MORE_WANT &&
-			    curstate != STATE_EXPECT_HAVE &&
-			    curstate != STATE_EXPECT_DONE) {
+			    curstate != STATE_EXPECT_HAVE_OR_DONE) {
 				err = got_error_msg(GOT_ERR_BAD_PACKET,
 				    "unexpected flush packet received");
 				goto done;
@@ -828,20 +833,19 @@ serve_read(int infd, int outfd, int gotd_sock, const char *repo_path,
 
 			if (curstate == STATE_EXPECT_WANT ||
 			    curstate == STATE_EXPECT_MORE_WANT ||
-			    curstate == STATE_EXPECT_HAVE) {
+			    curstate == STATE_EXPECT_HAVE_OR_DONE) {
 				err = forward_flushpkt(&ibuf);
 				if (err)
 					goto done;
 			}
-			if (curstate == STATE_EXPECT_HAVE && !have_ack) {
+			if (curstate == STATE_EXPECT_HAVE_OR_DONE &&
+			    !have_ack) {
 				err = send_nak(outfd, chattygot);
 				if (err)
 					goto done;
 			}
 			if (curstate == STATE_EXPECT_MORE_WANT)
-				curstate = STATE_EXPECT_HAVE;
-			else
-				curstate = STATE_EXPECT_DONE;
+				curstate = STATE_EXPECT_HAVE_OR_DONE;
 		} else if (n >= 5 && strncmp(buf, "want ", 5) == 0) {
 			if (curstate != STATE_EXPECT_WANT &&
 			    curstate != STATE_EXPECT_MORE_WANT) {
@@ -856,22 +860,18 @@ serve_read(int infd, int outfd, int gotd_sock, const char *repo_path,
 			if (curstate == STATE_EXPECT_WANT)
 				curstate = STATE_EXPECT_MORE_WANT;
 		} else if (n >= 5 && strncmp(buf, "have ", 5) == 0) {
-			if (curstate != STATE_EXPECT_HAVE &&
-			    curstate != STATE_EXPECT_DONE) {
+			if (curstate != STATE_EXPECT_HAVE_OR_DONE) {
 				err = got_error_msg(GOT_ERR_BAD_PACKET,
 				    "unexpected 'have' packet");
 				goto done;
 			}
-			if (curstate == STATE_EXPECT_HAVE) {
-				err = recv_have(&have_ack, outfd, &ibuf,
-				    buf, n, chattygot);
-				if (err)
-					goto done;
-				seen_have = 1;
-			}
+			err = recv_have(&have_ack, outfd, &ibuf,
+			    buf, n, chattygot);
+			if (err)
+				goto done;
+			seen_have = 1;
 		} else if (n == 5 && strncmp(buf, "done\n", 5) == 0) {
-			if (curstate != STATE_EXPECT_HAVE &&
-			    curstate != STATE_EXPECT_DONE) {
+			if (curstate != STATE_EXPECT_HAVE_OR_DONE) {
 				err = got_error_msg(GOT_ERR_BAD_PACKET,
 				    "unexpected 'done' packet");
 				goto done;
@@ -1268,7 +1268,8 @@ serve_write(int infd, int outfd, int gotd_sock, const char *repo_path,
 	while (curstate != STATE_EXPECT_PACKFILE) {
 		int n;
 		buf[0] = '\0';
-		err = got_pkt_readpkt(&n, infd, buf, sizeof(buf), chattygot);
+		err = got_pkt_readpkt(&n, infd, buf, sizeof(buf), chattygot,
+		    timeout);
 		if (err)
 			goto done;
 		if (n == 0) {
