@@ -44,6 +44,8 @@ test_file_changed() {
 	(printf "220\r\n250\r\n250\r\n250\r\n354\r\n250\r\n221\r\n" \
 		| timeout 5 nc -l "$GOTD_TEST_SMTP_PORT" > $testroot/stdout) &
 
+	sleep 1 # server starts up
+
 	got send -b main -q -r $testroot/repo-clone
 	ret=$?
 	if [ $ret -ne 0 ]; then
@@ -54,6 +56,7 @@ test_file_changed() {
 
 	wait %1 # wait for nc -l
 
+	short_commit_id=`trim_obj_id 28 $commit_id`
 	HOSTNAME=`hostname`
 	printf "HELO localhost\r\n" > $testroot/stdout.expected
 	printf "MAIL FROM:<${GOTD_USER}@${HOSTNAME}>\r\n" \
@@ -63,7 +66,7 @@ test_file_changed() {
 	printf "From: ${GOTD_USER}@${HOSTNAME}\r\n" >> $testroot/stdout.expected
 	printf "To: ${GOTD_DEVUSER}\r\n" >> $testroot/stdout.expected
 	printf "Subject: $GOTD_TEST_REPO_NAME: " >> $testroot/stdout.expected
-	printf "${GOTD_DEVUSER} changed refs/heads/main\r\n" \
+	printf "${GOTD_DEVUSER} changed refs/heads/main: $short_commit_id\r\n" \
 		>> $testroot/stdout.expected
 	printf "\r\n" >> $testroot/stdout.expected
 	printf "commit $commit_id\n" >> $testroot/stdout.expected
@@ -123,6 +126,8 @@ test_many_commits_not_summarized() {
 	(printf "220\r\n250\r\n250\r\n250\r\n354\r\n250\r\n221\r\n" \
 		| timeout 5 nc -l "$GOTD_TEST_SMTP_PORT" > $testroot/stdout) &
 
+	sleep 1 # server starts up
+
 	got send -b main -q -r $testroot/repo-clone
 	ret=$?
 	if [ $ret -ne 0 ]; then
@@ -133,6 +138,7 @@ test_many_commits_not_summarized() {
 
 	wait %1 # wait for nc -l
 
+	short_commit_id=`trim_obj_id 28 $commit_id`
 	HOSTNAME=`hostname`
 	printf "HELO localhost\r\n" > $testroot/stdout.expected
 	printf "MAIL FROM:<${GOTD_USER}@${HOSTNAME}>\r\n" \
@@ -143,7 +149,7 @@ test_many_commits_not_summarized() {
 		>> $testroot/stdout.expected
 	printf "To: ${GOTD_DEVUSER}\r\n" >> $testroot/stdout.expected
 	printf "Subject: $GOTD_TEST_REPO_NAME: " >> $testroot/stdout.expected
-	printf "${GOTD_DEVUSER} changed refs/heads/main\r\n" \
+	printf "${GOTD_DEVUSER} changed refs/heads/main: $short_commit_id\r\n" \
 		>> $testroot/stdout.expected
 	printf "\r\n" >> $testroot/stdout.expected
 	for i in `seq 1 24`; do
@@ -209,6 +215,8 @@ test_many_commits_summarized() {
 	(printf "220\r\n250\r\n250\r\n250\r\n354\r\n250\r\n221\r\n" \
 		| timeout 5 nc -l "$GOTD_TEST_SMTP_PORT" > $testroot/stdout) &
 
+	sleep 1 # server starts up
+
 	got send -b main -q -r $testroot/repo-clone
 	ret=$?
 	if [ $ret -ne 0 ]; then
@@ -219,6 +227,7 @@ test_many_commits_summarized() {
 
 	wait %1 # wait for nc -l
 
+	short_commit_id=`trim_obj_id 28 $commit_id`
 	HOSTNAME=`hostname`
 	printf "HELO localhost\r\n" > $testroot/stdout.expected
 	printf "MAIL FROM:<${GOTD_USER}@${HOSTNAME}>\r\n" \
@@ -229,7 +238,7 @@ test_many_commits_summarized() {
 		>> $testroot/stdout.expected
 	printf "To: ${GOTD_DEVUSER}\r\n" >> $testroot/stdout.expected
 	printf "Subject: $GOTD_TEST_REPO_NAME: " >> $testroot/stdout.expected
-	printf "${GOTD_DEVUSER} changed refs/heads/main\r\n" \
+	printf "${GOTD_DEVUSER} changed refs/heads/main: $short_commit_id\r\n" \
 		>> $testroot/stdout.expected
 	printf "\r\n" >> $testroot/stdout.expected
 	for i in `seq 1 51`; do
@@ -284,6 +293,8 @@ test_branch_created() {
 	(printf "220\r\n250\r\n250\r\n250\r\n354\r\n250\r\n221\r\n" \
 		| timeout 5 nc -l "$GOTD_TEST_SMTP_PORT" > $testroot/stdout) &
 
+	sleep 1 # server starts up
+
 	got send -b newbranch -q -r $testroot/repo-clone
 	ret=$?
 	if [ $ret -ne 0 ]; then
@@ -294,6 +305,7 @@ test_branch_created() {
 
 	wait %1 # wait for nc -l
 
+	short_commit_id=`trim_obj_id 28 $commit_id`
 	HOSTNAME=`hostname`
 	printf "HELO localhost\r\n" > $testroot/stdout.expected
 	printf "MAIL FROM:<${GOTD_USER}@${HOSTNAME}>\r\n" \
@@ -303,7 +315,7 @@ test_branch_created() {
 	printf "From: ${GOTD_USER}@${HOSTNAME}\r\n" >> $testroot/stdout.expected
 	printf "To: ${GOTD_DEVUSER}\r\n" >> $testroot/stdout.expected
 	printf "Subject: $GOTD_TEST_REPO_NAME: " >> $testroot/stdout.expected
-	printf "${GOTD_DEVUSER} created refs/heads/newbranch\r\n" \
+	printf "${GOTD_DEVUSER} created refs/heads/newbranch: $short_commit_id\r\n" \
 		>> $testroot/stdout.expected
 	printf "\r\n" >> $testroot/stdout.expected
 	printf "commit $commit_id\n" >> $testroot/stdout.expected
@@ -346,7 +358,10 @@ test_branch_removed() {
 	(printf "220\r\n250\r\n250\r\n250\r\n354\r\n250\r\n221\r\n" \
 		| timeout 5 nc -l "$GOTD_TEST_SMTP_PORT" > $testroot/stdout) &
 
+	sleep 1 # server starts up
+
 	local commit_id=`git_show_branch_head $testroot/repo-clone newbranch`
+	local short_commit_id=`trim_obj_id 28 $commit_id`
 
 	got send -d newbranch -q -r $testroot/repo-clone
 	ret=$?
@@ -367,7 +382,7 @@ test_branch_removed() {
 	printf "From: ${GOTD_USER}@${HOSTNAME}\r\n" >> $testroot/stdout.expected
 	printf "To: ${GOTD_DEVUSER}\r\n" >> $testroot/stdout.expected
 	printf "Subject: $GOTD_TEST_REPO_NAME: " >> $testroot/stdout.expected
-	printf "${GOTD_DEVUSER} removed refs/heads/newbranch\r\n" \
+	printf "${GOTD_DEVUSER} removed refs/heads/newbranch: $short_commit_id\r\n" \
 		>> $testroot/stdout.expected
 	printf "\r\n" >> $testroot/stdout.expected
 	printf "Removed refs/heads/newbranch: $commit_id\n" \
@@ -402,10 +417,15 @@ test_tag_created() {
 	got tag -r $testroot/repo-clone -m "new tag" 1.0 > /dev/null
 	local commit_id=`git_show_head $testroot/repo-clone`
 	local tagger_time=`git_show_tagger_time $testroot/repo-clone 1.0`
+	local tag_id=`got ref -r $testroot/repo-clone -l \
+		| grep "^refs/tags/$tag" | tr -d ' ' | cut -d: -f2`
+	local short_tag_id=`trim_obj_id 28 $tag_id`
 
 	(printf "220\r\n250\r\n250\r\n250\r\n354\r\n250\r\n221\r\n" \
 		| timeout 5 nc -l "$GOTD_TEST_SMTP_PORT" > $testroot/stdout) &
 
+	sleep 1 # server starts up
+
 	got send -t 1.0 -q -r $testroot/repo-clone
 	ret=$?
 	if [ $ret -ne 0 ]; then
@@ -425,7 +445,7 @@ test_tag_created() {
 	printf "From: ${GOTD_USER}@${HOSTNAME}\r\n" >> $testroot/stdout.expected
 	printf "To: ${GOTD_DEVUSER}\r\n" >> $testroot/stdout.expected
 	printf "Subject: $GOTD_TEST_REPO_NAME: " >> $testroot/stdout.expected
-	printf "${GOTD_DEVUSER} created refs/tags/1.0\r\n" \
+	printf "${GOTD_DEVUSER} created refs/tags/1.0: $short_tag_id\r\n" \
 		>> $testroot/stdout.expected
 	printf "\r\n" >> $testroot/stdout.expected
 	printf "tag refs/tags/1.0\n" >> $testroot/stdout.expected
@@ -478,10 +498,15 @@ test_tag_changed() {
 	got ref -r $testroot/repo-clone -d refs/tags/1.0 >/dev/null
 	got tag -r $testroot/repo-clone -m "new tag" 1.0 > /dev/null
 	local tagger_time=`git_show_tagger_time $testroot/repo-clone 1.0`
+	local tag_id=`got ref -r $testroot/repo-clone -l \
+		| grep "^refs/tags/$tag" | tr -d ' ' | cut -d: -f2`
+	local short_tag_id=`trim_obj_id 28 $tag_id`
 
 	(printf "220\r\n250\r\n250\r\n250\r\n354\r\n250\r\n221\r\n" \
 		| timeout 5 nc -l "$GOTD_TEST_SMTP_PORT" > $testroot/stdout) &
 
+	sleep 1 # server starts up
+
 	got send -f -t 1.0 -q -r $testroot/repo-clone
 	ret=$?
 	if [ $ret -ne 0 ]; then
@@ -501,7 +526,7 @@ test_tag_changed() {
 	printf "From: ${GOTD_USER}@${HOSTNAME}\r\n" >> $testroot/stdout.expected
 	printf "To: ${GOTD_DEVUSER}\r\n" >> $testroot/stdout.expected
 	printf "Subject: $GOTD_TEST_REPO_NAME: " >> $testroot/stdout.expected
-	printf "${GOTD_DEVUSER} changed refs/tags/1.0\r\n" \
+	printf "${GOTD_DEVUSER} changed refs/tags/1.0: $short_tag_id\r\n" \
 		>> $testroot/stdout.expected
 	printf "\r\n" >> $testroot/stdout.expected
 	printf "tag refs/tags/1.0\n" >> $testroot/stdout.expected
